from django.contrib import admin

from .models import Message, MessageLog

class MessageLogInline(admin.StackedInline):
    model = MessageLog
    extra = 0

class MessageAdmin(admin.ModelAdmin):
    inlines = [MessageLogInline,]

admin.site.register(Message, MessageAdmin)


from django.conf import settings

from .service import compose, message_handler, handle_message

def init():
    for app in settings.INSTALLED_APPS:
        try:
            __import__(app + '.ymessage_handlers')
        except ImportError:
            pass

init()


# Generated by Django 4.2.5 on 2023-10-07 08:17

from django.db import migrations, models
import django.db.models.deletion
import ool


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False)),
                ('type', models.CharField(max_length=50)),
                ('content', models.JSONField(blank=True, null=True)),
                ('status', models.CharField(choices=[('queuing', 'queuing'), ('processing', 'processing'), ('sent', 'sent'), ('exception', 'exception'), ('no_handler', 'no_handler')], default='queuing', max_length=20)),
                ('update_at', models.DateTimeField(auto_now=True)),
                ('create_at', models.DateTimeField(auto_now_add=True)),
                ('version', ool.VersionField(default=0)),
            ],
            options={
                'ordering': ('-create_at',),
                'get_latest_by': 'create_at',
            },
            bases=(ool.VersionedMixin, models.Model),
        ),
        migrations.CreateModel(
            name='MessageLog',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('content', models.JSONField(blank=True, null=True)),
                ('create_at', models.DateTimeField(auto_now_add=True)),
                ('message', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ymessage.message')),
            ],
            options={
                'ordering': ('id',),
            },
        ),
        migrations.AddIndex(
            model_name='message',
            index=models.Index(fields=['create_at'], name='idx_message__create_at'),
        ),
    ]



import json

from django.http import HttpResponse
from . import service

def default_message_handler(request):
    sleep = 0
    try:
        service.handle_message()
    except service.NoMoreMessages:
        sleep = 3

    return HttpResponse(json.dumps({
        'sleep': sleep
    }), content_type='application/json')


from django.db import models

from ool import VersionField, VersionedMixin

class Message(VersionedMixin, models.Model):
    STATUS_QUEUING = 'queuing'
    STATUS_PROCESSING = 'processing'
    STATUS_SENT = 'sent'
    STATUS_EXCEPTION = 'exception'
    STATUS_NO_HANDLER = 'no_handler'
    STATUS_CHOICE = (
        (STATUS_QUEUING, STATUS_QUEUING),
        (STATUS_PROCESSING, STATUS_PROCESSING),
        (STATUS_SENT, STATUS_SENT),
        (STATUS_EXCEPTION, STATUS_EXCEPTION),
        (STATUS_NO_HANDLER, STATUS_NO_HANDLER),
    )
    id = models.UUIDField(primary_key=True)
    type = models.CharField(max_length=50)
    content = models.JSONField(null=True, blank=True)
    status = models.CharField(max_length=20, choices=STATUS_CHOICE, default=STATUS_QUEUING)
    update_at = models.DateTimeField(auto_now=True)
    create_at = models.DateTimeField(auto_now_add=True)
    version = VersionField()

    class Meta:
        get_latest_by = 'create_at'
        ordering = ('-create_at',)
        indexes = [
            models.Index(fields=["create_at"], name='idx_message__create_at'),
        ]

    def log(self, content):
        return MessageLog.objects.create(content=content, message=self)


    def begin_processing(self):
        self.status = Message.STATUS_PROCESSING
        self.save()


    def sent(self):
        self.status = self.STATUS_SENT
        self.save()

    def catch_exception(self, reason):
        self.log({
            'exception': reason
        })
        self.status = self.STATUS_EXCEPTION
        self.save()

    def no_handler(self):
        self.status = self.STATUS_NO_HANDLER
        self.save()

    def is_processing(self):
        return self.status == self.STATUS_PROCESSING

    def __str__(self):
        return f"{self.create_at} {self.type}:{self.id} {self.status}"


class MessageLog(models.Model):
    message = models.ForeignKey(Message, on_delete=models.CASCADE)
    content = models.JSONField(null=True, blank=True)
    create_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ('id', )


import traceback
import ool

all_handlers = {}

types_for_all_handlers = []

def add_handler(type, handler):
    all_handlers[type] = handler
    types_for_all_handlers.append(type)


def message_handler(type):
    def add_handler_of_type(func):
        add_handler(type, func)
        return func 
    return add_handler_of_type


def compose(type, id, content=None):
    from .models import Message
    return Message.objects.create(type=type,
                                  id=id,
                                  content=content,
                                  status=Message.STATUS_QUEUING)


class NoMoreMessages(Exception):
    pass


def get_next_message_for_processing(types):
    from .models import Message

    message = None
    while message is None:
        message = Message.objects.filter(status=Message.STATUS_QUEUING, type__in=types).order_by('create_at').first()
        if message is not None:
            try:
                message.begin_processing()
            except ool.ConcurrentUpdate:
                message = None
        else:
            raise NoMoreMessages()
    return message


def handle_message():
    message = get_next_message_for_processing(types_for_all_handlers)
    try:
        handler = all_handlers.get(message.type, None)
        if handler == None:
            message.no_handler()
        else:
            handler(message)
    except:
        message.catch_exception({
            'trace': traceback.format_exc()
        })

        
